
package cz.cuni.mff.corrector.execution;

import java.util.ArrayList;
import java.util.List;

/**
 * Container for scheduling dependencies of a particular associated task.
 * 
 * Note that instances of this class only serve as containers for data.
 * It is the responsibility of callers to synchronise correctly on given
 * associated task instances in order to perform more complex operations
 * correctly. Isolated synchronisation over public methods this container
 * would not be enough!
 * 
 * @author Martin Svoboda
 */
public class Dependencies {
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Definition of all possible task execution states.
     * 
     * The only purpose of these states is to avoid race conditions under
     * which the execution of the same nested task would be requested
     * by several parental tasks at the same time.
     */
    public enum State {
        /**
         * Task execution state: activated (requested execution).
         */
        ACTIVATED,
        /**
         * Task execution state: disabled (not requested by anyone).
         */
        DISABLED;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Current value of the execution state of the task.
     */
    private State executionState;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Set of tasks that have requested execution of this particular task.
     */
    private List<Task> requestingTasks;

    /**
     * Set of tasks that were requested by this particular task.
     */
    private List<Task> requestedTasks;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates a new instance of the dependencies container.
     * @see Task#createExtendedTask
     */
    protected Dependencies() {
    
        // Initialising internal variables
        this.executionState = State.DISABLED;
        this.requestingTasks = new ArrayList<Task>();
        this.requestedTasks = new ArrayList<Task>();
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the current value of the the execution state.
     * Note that the caller is responsible for the appropriate synchronisation!
     * @return Current value of the task execution state.
     */
    public State getExecutionState() {
        return this.executionState;
    }
    
    /**
     * Detects whether the execution state is currently activated.
     * Note that the caller is responsible for the appropriate synchronisation!
     * @return Returns <code>true</code> if the task is currently activated.
     */
    public boolean hasStateActivated() {
        return (this.executionState == State.ACTIVATED);
    }
    
    /**
     * Detects whether the execution state is currently disabled.
     * Note that the caller is responsible for the appropriate synchronisation!
     * @return Returns <code>true</code> if the task is currently disabled.
     */
    public boolean hasStateDisabled() {
        return (this.executionState == State.DISABLED);
    }

    /**
     * Sets a new value of the execution state of the task.
     * Note that the caller is responsible for the appropriate synchronisation!
     * @param newState New value of the task execution state.
     */
    public void setExecutionState(State newState) {
        this.executionState = newState;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the set of tasks that have requested execution of this task.
     * Despite the reference to the internal container is returned,
     * the caller is not allowed to directly manipulated with it.
     * Note that the caller is responsible for the appropriate synchronisation!
     * @return Set of tasks waiting for this task.
     */
    public List<Task> getRequestingTasks() {
        return this.requestingTasks;
    }
    
    /**
     * Fetches some requesting task from the internal container.
     * Note that the caller is responsible for the appropriate synchronisation!
     * @return Returns reference to some requesting task.
     */
    public Task getRequestingTask() {
        return this.requestingTasks.get(0);
    }
    
    /**
     * Adds a new requesting task into the internal container.
     * Note that the caller is responsible for the appropriate synchronisation!
     * @param requestingTask Reference to the task to be added.
     */
    public void addRequestingTask(Task requestingTask) {
        this.requestingTasks.add(requestingTask);
    }
    
    /**
     * Removes the requesting task from the internal container.
     * Note that the caller is responsible for the appropriate synchronisation!
     * @param requestingTask Reference to the task to be removed.
     */
    public void removeRequestingTask(Task requestingTask) {
        this.requestingTasks.remove(requestingTask);
    }
    
    /**
     * Detects whether the task has at least one requesting task.
     * Note that the caller is responsible for the appropriate synchronisation!
     * @return Returns <code>true</code> if there exists some requesting tasks.
     */
    public boolean hasRequestingTasks() {
        return (!this.requestingTasks.isEmpty());
    }
    
    
    // -------------------------------------------------------------------------
    

    /**
     * Returns the set of tasks requested by this task.
     * Despite the reference to the internal container is returned,
     * the caller is not allowed to directly manipulated with it.
     * Note that the caller is responsible for the appropriate synchronisation!
     * @return Set of tasks for which this task is waiting.
     */
    public List<Task> getRequestedTasks() {
        return this.requestedTasks;
    }
    
    /**
     * Adds a new requesting task into the internal container.
     * Note that the caller is responsible for the appropriate synchronisation!
     * @param requestedTask Reference to the task to be added.
     */
    public void addRequestedTask(Task requestedTask) {
        this.requestedTasks.add(requestedTask);
    }
    
    /**
     * Removes the requested task from the internal container.
     * Note that the caller is responsible for the appropriate synchronisation!
     * @param requestedTask Reference to the task to be removed.
     */
    public void removeRequestedTask(Task requestedTask) {
        this.requestedTasks.remove(requestedTask);
    }
    
    /**
     * Detects whether the task has at least one requested task.
     * Note that the caller is responsible for the appropriate synchronisation!
     * @return Returns <code>true</code> if there exists some requested tasks.
     */
    public boolean hasRequestedTasks() {
        return (!this.requestedTasks.isEmpty());
    }
    
    /**
     * Detects whether there are not requested tasks.
     * Note that the caller is responsible for the appropriate synchronisation!
     * @return Returns <code>true</code> if there exist no requested tasks.
     */
    public boolean hasNoRequestedTasks() {
        return (this.requestedTasks.isEmpty());
    }
    
    
    // -------------------------------------------------------------------------
    
}
