
package cz.cuni.mff.corrector.execution;

import cz.cuni.mff.corrector.intent.Intent;
import cz.cuni.mff.corrector.repair.Repair;

/**
 * Task object describing state of the correction intent processing.
 * 
 * @author Martin Svoboda
 */
public class Task {
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Reference to the associated correction intent.
     */
    private final Intent correctionIntent;
    
    /**
     * Reference to the associated intent repair structure.
     * Note that an associated repair is always and automatically created
     * and registered during the task instance construction. However, such
     * repair is not yet assigned a sequence repair (and, thus, correction
     * multigraph) and the completeness flag is set to false.
     * When the task starts its execution for the first time, the given
     * correction multigraph is created and registered. On the other hand,
     * function for resolving the intent repair cost is working correctly
     * even before this multigraph registration (and only focuses on node
     * repair and its cost)!
     * Once the repair is fully computed and the task attained its goal,
     * the corresponding completeness flag of the repair structure has
     * to be marked as active.
     * This is important since the handler and corrector synchronisation
     * relies on this fact!
     * @see Repair#isCompletelyEvaluated
     */
    protected Repair intentRepair;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Optional reference to the associated working variables.
     */
    private Variables workingVariables;
    
    
    /**
     * Optional reference to the associated execution dependencies.
     */
    private Dependencies executionDeps;
    
    
    /**
     * Optional value of the scheduling management parameter.
     * This parameter is only used in case of the following approaches:
     *   SCHEDULING_SINGLE and SCHEDULING_MULTIPLE.
     * The starting correction intent should be marked as the master
     * (i.e. <code>true</code>) since after it is completely evaluated
     * it cannot notify its parents (there are no ones), but must manually
     * notify the thread that initiated the entire correction.
     */
    private Boolean schedulingMaster;
    
    
    /**
     * Optional reference to the parental requesting task.
     * This value is only relevant in case of the SCHEDULING_SINGLE approach.
     * It is used to maintain the calling hierarchy and back notifications.
     */
    private Task requestingTask;
    
    
    /**
     * Data container for managing requested tasks and postponed intents.
     * This parameter is only used in case of the following approaches:
     *   SCHEDULING_SINGLE and SCHEDULING_MULTIPLE.
     */
    private Requests requestsContainer;
    
    
    /**
     * Description of granted and used quota for the iterative execution.
     * Only used in case of the ITERATIVE algorithm strategy.
     */
    private Quota quotaDescription;
    
    
    /**
     * Internal flag for marking completed refinement request.
     * This flag is set to <code>false</code> when a new quota is set
     * and to <code>true</code> before the parent is notified.
     */
    private boolean completedJob;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Internal constructor for a new task instance.
     * Note that an intent repair is always created as well!
     * However, the multigraph for this repair must be registered later on.
     * @param correctionIntent Associated correction intent reference.
     */
    private Task(Intent correctionIntent) {
        
        // Initialising intent components
        this.correctionIntent = correctionIntent;
        this.intentRepair = new Repair(correctionIntent);
        
        // Initialising working components
        this.workingVariables = null;
        this.executionDeps = null;
        this.schedulingMaster = null;
        this.requestingTask = null;
        this.requestsContainer = null;
        this.quotaDescription = null;
        this.completedJob = false;
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates a new simple task that only holds the intent reference.
     * Working variables must be created and associated explicitly.
     * @param correctionIntent Associated correction intent reference.
     * @return Newly created and initialised task instance.
     */
    public static Task createSimpleTask(Intent correctionIntent) {
        
        // Creating a new task instance
        Task newTask = new Task(correctionIntent);
        
        // Returning the task reference
        return newTask;
        
    }
    
    /**
     * Creates a new extended task with initialised dependencies.
     * Working variables must be created and associated explicitly.
     * @param correctionIntent Associated correction intent reference.
     * @return Newly created and initialised task instance.
     */
    public static Task createExtendedTask(Intent correctionIntent) {
        
        // Creating a new task instance
        Task newTask = new Task(correctionIntent);
        
        // Initialising task dependencies
        newTask.executionDeps = new Dependencies();
        
        // Returning the task reference
        return newTask;
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the associated correction intent.
     * IMPORTANT NOTICE: Note that when signature handling is enabled,
     * this correction intent reference is fully valid and makes sense
     * only within the task itself and especially not from other tasks
     * that would only like to wait until this repair is available!
     * @return Reference to the associated correction intent.
     */
    public Intent getCorrectionIntent() {
        return this.correctionIntent;
    }
    
    
    /**
     * Returns the reference to the computed intent repair.
     * Note that the returned repair may not yet be fully evaluated!
     * Therefore the detection of its completed evaluation has to be based
     * on the completeness flag maintained directly within the repair.
     * @see #hasCompletedRapair
     * @see Repair#isCompletelyEvaluated
     * @return Reference to the intent repair structure.
     */
    public Repair getIntentRepair() {
        return this.intentRepair;
    }
    
    /**
     * Detects whether the registered repair is already fully evaluated.
     * @return Returns <code>true</code> in case the repair is registered
     * and at the same time already completely evaluated.
     */
    public boolean hasCompletedRapair() {
        return
            (this.intentRepair != null) &&
            (this.intentRepair.isCompletelyEvaluated());
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the associated working variables.
     * @return Reference to the associated working variables container.
     */
    public Variables getWorkingVariables() {
        return this.workingVariables;
    }
    
    /**
     * Associates the working variables container with the task.
     * @param workingVariables Working variables container to be registered.
     */
    public void setWorkingVariables(Variables workingVariables) {
        this.workingVariables = workingVariables;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the internal execution dependencies container.
     * @return Reference to the execution dependencies.
     */
    public Dependencies getExecutionDependencies() {
        return this.executionDeps;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the current value of the scheduling management parameter.
     * Note that this function is only relevant for scheduling approaches.
     * @return Returns <code>true</code> for master scheduling behavior.
     */
    public Boolean getSchedulingMaster() {
        return this.schedulingMaster;
    }
    
    /**
     * Sets the scheduling management parameter.
     * Note that this function is only relevant for scheduling approaches.
     * Also note that this parameter has to be only set during
     * parent-child activations and not child-parent ones!
     * @param newValue <code>True</code> for master scheduling behavior.
     * 
     * If and only if task itself is set as master, the requesting task
     * reference must be set to <code>null</code>.
     * @see #setRequestingTask
     */
    public void setSchedulingMaster(Boolean newValue) {
        this.schedulingMaster = newValue;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the parental requesting task.
     * This method is only relevant for the SCHEDULING_SINGLE approach.
     * @return Reference to the requesting task in the call hierarchy.
     */
    public Task getRequestingTask() {
        return this.requestingTask;
    }
    
    /**
     * Sets the reference of the parental requesting task.
     * This method is only relevant for the SCHEDULING_SINGLE approach.
     * @param parentalTask Reference to the requesting task in the hierarchy.
     * 
     * If and only if this requesting task reference is set <code>null</code>,
     * this task itself must be set as master.
     * @see #setSchedulingMaster
     */
    public void setRequestingTask(Task parentalTask) {
        this.requestingTask = parentalTask;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the data container for managing scheduling requests.
     * Note that the container is not created automatically.
     * @return Reference to the internal requests container.
     */
    public Requests getRequestsContainer() {
        return this.requestsContainer;
    }
    
    /**
     * Creates a new requests container and stores its reference.
     */
    public void createRequestsContainer() {
        this.requestsContainer = new Requests();
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Sets a new value for the entire iterative quota description.
     * Also note that the flag of completed refinement job is reset.
     * @param newQuota Quota object to be registered with the task.
     */
    public void setQuotaData(Quota newQuota) {
        this.quotaDescription = newQuota;
        this.completedJob = false;
    }
    
    /**
     * Returns the description object of the iterative quota.
     * @return Reference to the registered quota or <code>null</code>.
     */
    public Quota getQuotaData() {
        return this.quotaDescription;
    }
    
    
    /**
     * Activates the internal flag of completed refinement job.
     */
    public void markCompletedJob() {
        this.completedJob = true;
    }
    
    /**
     * Detects whether the requested refinement was already completed.
     * @return Returns the current value of the refinement flag.
     */
    public boolean hasCompletedJob() {
        return this.completedJob;
    }
    
    
    // -------------------------------------------------------------------------
    
}
