
package cz.cuni.mff.corrector.execution;

import cz.cuni.mff.corrector.graph.Edge;
import cz.cuni.mff.corrector.graph.Graph;
import cz.cuni.mff.corrector.graph.Vertex;
import cz.cuni.mff.corrector.intent.Intent;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Working variables container to be used with all correction strategies.
 * This container should be created and associated with a particular task.
 * 
 * @author Martin Svoboda
 */
public class Variables {
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Definition of all possible internal states of the handler execution.
     */
    public enum Phase {
        /**
         * Default handler phase: newly created and not yet initialised.
         * This is the default phase associated with a variables container.
         */
        CREATED,
        /**
         * Default handler phase: initialised and running exploration.
         */
        EXPLORING,
        /**
         * Default handler phase: exploration finished and searching paths.
         */
        EXPLORED,
        /**
         * Default handler phase: repair composed and completely finished.
         */
        FINISHED;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Current value of the internal handler processing phase.
     */
    private Phase executionPhase;
    
    
    /**
     * Associated correction multigraph.
     */
    private Graph correctionGraph;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * List of reached vertices to be further explored and processed.
     * This list is used during the graph exploration phase.
     */
    private Set<Vertex> reachedVertices;
    
    
    /**
     * Optional value of the resolved total correction cost.
     * Only relevant for EXPLORING and ITERATIVE strategies.
     * This value is used when exploring the multigraph in order to correctly
     * stop the search for the shortest paths. It means that the variable
     * is initialized to <code>null</code> and when the first target vertex
     * is reach at the last stratum, the current reached cost value is
     * declared as the final cost and stored in this attribute.
     */
    private Integer fixedCost;
    
    
    /**
     * Container for storing delayed and not yet closed edges.
     * Only relevant for all approaches of the ITERATIVE strategy.
     * This container is something else than prepared edges in requests.
     */
    private List<Edge> delayedEdges;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Optional value of the current reached correction cost.
     * Only relevant for the ITERATIVE strategy.
     * This variable holds the reached cost to the last reached vertex
     * that was selected to be processed, regardless it was complete or
     * not and regardless we could also already even exceed the correctly
     * resolved final total correction cost.
     */
    private Integer reachedCost;
    
    
    /**
     * Internal caching map for all task for reached and not closed edges.
     * Only relevant for all approaches of the ITERATIVE strategy.
     * Is allows to maintain references to all required nested tasks.
     * Note that this mechanism is required, since not all approaches
     * are able to rely on global tasks caching.
     */
    private Map<Intent, Task> tasksCache;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates and initializes a new instance of the variables container.
     * Note that the correction multigraph instance is not initialized.
     * Also note that only the core and default components are prepared!
     * @see #setCorrectionGraph
     */
    public Variables() {
        
        // Initialising the execution state and graph reference
        initCoreComponents();
        
        // Initialising default variables for graph exploration
        initDefaultComponents();
        
        // Iterative components must be initialized manually
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Initializes the core components of the variables container.
     * These are the default execution state and graph reference.
     * This method should only be called only from the constructor.
     */
    private void initCoreComponents() {
        this.executionPhase = Phase.CREATED;
        this.correctionGraph = null;
    }
    
    /**
     * Initializes the default components of the variables container.
     * These are variables required to the standard graph exploration.
     * This method should only be called only from the constructor.
     */
    private void initDefaultComponents() {
        this.reachedVertices = new HashSet<Vertex>();
        this.reachedCost = null;
        this.fixedCost = null;
    }
    
    /**
     * Initializes the iterative components of the variables container.
     * These are containers for storing open and closed reached minimal
     * costs, delayed edges and local cache of tasks for reached edges.
     */
    public void initIterativeComponents() {
        this.delayedEdges = new ArrayList<Edge>();
        this.tasksCache = new HashMap<Intent, Task>();
    }
    
    
    /**
     * Clears default and iterative components but preserves the core ones.
     * This method should be called when to goal of the task processing
     * was already attained and there is no need to preserve working data
     * any longer except the execution phase flag that might be required.
     * Note that the internal components become unusable, not just cleared!
     */
    public void clearVariablesContainer() {
        this.reachedVertices = null;
        this.reachedCost = null;
        this.fixedCost = null;
        this.delayedEdges = null;
        this.tasksCache = null;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the current value of the internal execution phase.
     * @return Handler execution phase.
     */
    public Phase getExecutionPhase() {
        return this.executionPhase;
    }
    
    /**
     * Detects whether the execution phase corresponds to the given value.
     * @param queriedPhase Phase value to be tested.
     * @return Returns <code>true</code> if the phase corresponds.
     */
    public boolean hasExecutionPhase(Phase queriedPhase) {
        return (this.executionPhase == queriedPhase);
    }
    
    /**
     * Sets a new value of the internal execution phase.
     * @param newPhase New handler execution phase.
     */
    public void setExecutionPhase(Phase newPhase) {
        this.executionPhase = newPhase;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the associated correction multigraph.
     * @return Reference to the correction multigraph structure.
     */
    public Graph getCorrectionGraph() {
        return this.correctionGraph;
    }
    
    /**
     * Registers the provided correction multigraph with the container.
     * @param correctionGraph Reference to the graph to be registered.
     */
    public void setCorrectionGraph(Graph correctionGraph) {
        this.correctionGraph = correctionGraph;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Inserts the given vertex into the list of reached vertices.
     * @param reachedVertex Vertex to be added into the internal container.
     */
    public void addReachedVertex(Vertex reachedVertex) {
        this.reachedVertices.add(reachedVertex);
    }
    
    /**
     * Detects whether at least one reached vertex exists.
     * @return Returns <code>true</code> if reached vertices are not empty.
     */
    public boolean hasReachedVertices() {
        return (!this.reachedVertices.isEmpty());
    }
    
    /**
     * Detects whether the given vertex is stored in the reached vertices.
     * @param inspectedVertex Vertex to be inspected.
     * @return Returns <code>true</code> if the vertex is stored as reached.
     */
    public boolean hasReachedVertex(Vertex inspectedVertex) {
        return (this.reachedVertices.contains(inspectedVertex));
    }
    
    /**
     * Returns the first (some) reached vertex.
     * Note that this vertex is not removed from the container by this method.
     * Only suitable for the DEFAULT algorithm strategy!
     * The order of vertices in the internal container is not defined!
     * @return Reached vertex returned as a first item by the iterator.
     */
    public Vertex getReachedVertex() {
        return this.reachedVertices.iterator().next();
    }
    
    /**
     * Returns all the stored reached vertices.
     * @return Reference to the internal container of reached vertices.
     */
    public Set<Vertex> getReachedVertices() {
        return this.reachedVertices;
    }
    
    /**
     * Removes the given vertex from the list of reached vertices.
     * @param reachedVertex Vertex to be removed from the internal container.
     */
    public void removeReachedVertex(Vertex reachedVertex) {
        this.reachedVertices.remove(reachedVertex);
    }
    
    /**
     * Deletes all items from the list of reached vertices.
     */
    public void clearReachedVertices() {
        this.reachedVertices.clear();
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the current value of the correction final cost.
     * Only relevant for the ITERATIVE strategy.
     * @return Current value of the reached cost.
     * @see #reachedCost
     */
    public Integer getReachedCost() {
        return this.reachedCost;
    }
    
    
    /**
     * Sets the value of the reached correction cost.
     * Only relevant for the ITERATIVE strategy.
     * @param reachedCost New value of the currently reached cost.
     */
    public void setReachedCost(int reachedCost) {
        this.reachedCost = reachedCost;
    }
    
    
    /**
     * Returns the current value of the correction final cost.
     * @return Value of the final cost or <code>null</code> when not yet set.
     * @see #fixedCost
     */
    public Integer getFixedCost() {
        return this.fixedCost;
    }
    
    
    /**
     * Sets the value of the final correction cost.
     * @param fixedCost Final value of the multigraph correction.
     */
    public void setFixedCost(int fixedCost) {
        this.fixedCost = fixedCost;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * fetched the required task from the local cache if it exists.
     * @param nestedIntent Correction intent that defines the request.
     * @return Reference to the fetched task or <code>null</code>.
     */
    public Task getCachedTask(Intent nestedIntent) {
        return this.tasksCache.get(nestedIntent);
    }
    
    /**
     * Adds a new task into the local cache of tasks for reached intents.
     * @param nestedIntent Reference to the indexing intent.
     * @param nestedTask Reference to the task to be inserted.
     */
    public void addCachedTask(Intent nestedIntent, Task nestedTask) {
        this.tasksCache.put(nestedIntent, nestedTask);
    }
    
    /**
     * Removes the cached task for the specified correction intent.
     * @param nestedIntent Reference to the indexing intent.
     */
    public void removeCachedTask(Intent nestedIntent) {
        this.tasksCache.remove(nestedIntent);
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the internal container with delayed and not yet closed edges.
     * @return Reference to the working set of delayed edges.
     */
    public List<Edge> getDelayedEdges() {
        return this.delayedEdges;
    }
    
    
    /**
     * Adds a new edge to the container of delayed edges.
     * @param delayedEdge Edge to be added between delayed edges.
     */
    public void addDelayedEdge(Edge delayedEdge) {
        this.delayedEdges.add(delayedEdge);
    }
    
    /**
     * Adds all provided edges to the container of delayed edges.
     * @param delayedEdges Edges to be added between delayed edges.
     */
    public void addDelayedEdges(Collection<Edge> delayedEdges) {
        this.delayedEdges.addAll(delayedEdges);
    }
    
    
    /**
     * Removes all edges from the container of delayed edges.
     */
    public void removeDelayedEdges() {
        this.delayedEdges.clear();
    }
    
    /**
     * Removes the selected edge from the container of delayed edges.
     * @param delayedEdge Edge to be removed from delayed edges.
     */
    public void removeDelayedEdge(Edge delayedEdge) {
        this.delayedEdges.remove(delayedEdge);
    }
    
    
    /**
     * Detects whether there exists at least one delayed edge.
     * @return Returns <code>true</code> if there exist delayed edges.
     */
    public boolean hasDelayedEdges() {
        return (!this.delayedEdges.isEmpty());
    }
    
    
    // -------------------------------------------------------------------------
    
}
