
package cz.cuni.mff.corrector.expression;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * Regular subexpression for union operation.
 *
 * @author Martin Svoboda
 */
public class ExpressionUnion extends Expression {

    // -------------------------------------------------------------------------


    /**
     * Left subexpression of the union operation.
     */
    private Expression leftOperand;

    /**
     * Right subexpression of the union operation.
     */
    private Expression rightOperand;


    // -------------------------------------------------------------------------


    /**
     * Creates an instance of the union regular subexpression.
     * @param leftOperand Left subexpression of the concatenation operation.
     * @param rightOperand Right subexpression of the concatenation operation.
     */
    protected ExpressionUnion(
        Expression leftOperand,
        Expression rightOperand
    ) {
        this.leftOperand = leftOperand;
        this.rightOperand = rightOperand;
    }


    public ExpressionUnion duplicateExpression() {
        return new ExpressionUnion(
            this.leftOperand.duplicateExpression(),
            this.rightOperand.duplicateExpression()
        );
    }


    // -------------------------------------------------------------------------


    public Type getType() {
        return Type.UNION;
    }


    public String serializeExpression() {
        return
            TOKEN_MT_OPEN +
            this.leftOperand.serializeExpression() +
            TOKEN_OP_UNION +
            this.rightOperand.serializeExpression() +
            TOKEN_MT_CLOSE;
    }


    public Set<String> getSymbols() {
        Set<String> symbols = new HashSet<String>();
        symbols.addAll(leftOperand.getSymbols());
        symbols.addAll(rightOperand.getSymbols());
        return symbols;
    }


    protected void markSymbols(Provider provider) {
        this.leftOperand.markSymbols(provider);
        this.rightOperand.markSymbols(provider);
    }


    // -------------------------------------------------------------------------


    public boolean resolveEmpty() {
        if (this.allowsEmpty == null) {
            this.allowsEmpty =
                leftOperand.resolveEmpty() || rightOperand.resolveEmpty();
        }
        return this.allowsEmpty;
    }

    public Set<String> resolveFirst() {
        Set<String> symbols = new HashSet<String>(leftOperand.resolveFirst());
        symbols.addAll(rightOperand.resolveFirst());
        return symbols;
    }

    public Set<String> resolveLast() {
        Set<String> symbols = new HashSet<String>(rightOperand.resolveLast());
        symbols.addAll(leftOperand.resolveLast());
        return symbols;
    }

    public Map<String, Set<String>> resolveFollow() {
        Map<String, Set<String>> mapping = new HashMap<String, Set<String>>();
        Map<String, Set<String>> leftMap = leftOperand.resolveFollow();
        for (String keySymbol : leftMap.keySet()) {
            Set<String> newSymbols =
                new HashSet<String>(leftMap.get(keySymbol));
            mapping.put(keySymbol, newSymbols);
        }
        Map<String, Set<String>> rightMap = rightOperand.resolveFollow();
        for (String keySymbol : rightMap.keySet()) {
            if (mapping.containsKey(keySymbol)) {
                mapping.get(keySymbol).addAll(rightMap.get(keySymbol));
            } else {
                Set<String> newSymbols =
                    new HashSet<String>(rightMap.get(keySymbol));
                mapping.put(keySymbol, newSymbols);
            }
        }
        return mapping;
    }


    // -------------------------------------------------------------------------

}
