
package cz.cuni.mff.corrector.grammar;

import cz.cuni.mff.corrector.expression.Expression;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * Regular tree grammar representation.
 *
 * @author Martin Svoboda
 */
public class Grammar {

    // -------------------------------------------------------------------------


    /**
     * Set of all defined nonterminals.
     */
    private Set<String> nonterminalSymbols;

    /**
     * Set of all defined terminals.
     */
    private Set<String> terminalSymbols;

    /**
     * Set of all defined nonterminal starting symbols.
     */
    private Set<String> startingSymbols;


    /**
     * Set of all defined production rules.
     */
    private Set<Rule> productionRules;


    /**
     * Container for all defined grammar contexts.
     * Map is indexed using the nonterminal symbol first and then terminal.
     */
    private Map<String, Map<String, Context>> grammarContexts;


    /**
     * Reference to the special starting grammar rule.
     */
    private Rule startingRule;

    /**
     * Reference to the special starting grammar context.
     */
    private Context startingContext;


    /**
     * Reference to the special empty grammar rule.
     */
    private Rule emptyRule;

    /**
     * Reference to the special empty grammar context.
     */
    private Context emptyContext;
    

    // -------------------------------------------------------------------------


    /**
     * Creates a new and empty grammar object.
     */
    private Grammar() {
        this.nonterminalSymbols = new HashSet<String>();
        this.terminalSymbols = new HashSet<String>();
        this.startingSymbols = new HashSet<String>();
        this.productionRules = new HashSet<Rule>();
        this.grammarContexts = new HashMap<String, Map<String, Context>>();
        this.startingRule = null;
        this.startingContext = null;
        this.emptyRule = null;
        this.emptyContext = null;
    }


    /**
     * Initializes a new grammar from a provided consistent data string.
     * @param grammarString String with a grammar definition.
     * @return Completely prepared grammar object.
     */
    public static Grammar createFromDataString(String grammarString) {

        // Creating an empty grammar object
        Grammar grammar = new Grammar();

        // Splitting lines of the definition
        String[] linesArray = grammarString.split("\n");
        grammar.startingSymbols.addAll(
            Arrays.asList(linesArray[0].split(" +"))
        );
        // Processing production rules
        for (int row = 1; row < linesArray.length; row++) {
            String[] items = linesArray[row].split(" +");
            if (items.length < 3) {
                break;
            }
            Rule rule = new Rule(items[0], items[1], items[2]);
            grammar.productionRules.add(rule);
        }

        // Resolving sets of symbols
        for (Rule rule : grammar.productionRules) {
            grammar.nonterminalSymbols.add(rule.getNonterminalSymbol());
            grammar.terminalSymbols.add(rule.getTerminalSymbol());
        }

        // Creating a starting grammar context
        StringBuilder startingModel = new StringBuilder();
        for (String startingNonterminal : grammar.startingSymbols) {
            startingModel.append(startingNonterminal);
            startingModel.append(Expression.TOKEN_OP_UNION);
        }
        grammar.startingRule = new Rule(
            "#root",
            "#root",
            startingModel.substring(0, startingModel.length() - 1)
        );
        grammar.startingContext = new Context(grammar.startingRule, grammar);

        // Creating an empty grammar context
        grammar.emptyRule = new Rule(null, null, Expression.TOKEN_RE_EMPTY);
        grammar.emptyContext = new Context(grammar.emptyRule, grammar);

        // Creating general contexts for rules
        for (Rule rule : grammar.productionRules) {
            Context context = new Context(rule, grammar);
            // Inserting the context into the container
            String nonterminal = rule.getNonterminalSymbol();
            if (grammar.grammarContexts.containsKey(nonterminal)) {
                grammar.grammarContexts.get(nonterminal).put(
                    rule.getTerminalSymbol(),
                    context
                );
            } else {
                Map<String, Context> contexts = new HashMap<String, Context>();
                contexts.put(
                    rule.getTerminalSymbol(),
                    context
                );
                grammar.grammarContexts.put(nonterminal, contexts);
            }
        }

        // Returning initalizes grammar structure
        return grammar;

    }


    /**
     * Initializes a new grammar from a provided data file.
     * @param fileName Path to the file containing grammar definition.
     * @return Completely prepared grammar object.
     */
    public static Grammar createFromDataFile(String fileName) {
        try {
            FileReader fileReader = new FileReader(fileName);
            BufferedReader bufferedReader = new BufferedReader(fileReader);
            StringBuilder stringBuilder = new StringBuilder();
            String stringLine;
            while ((stringLine = bufferedReader.readLine()) != null) {
                stringBuilder.append(stringLine).append("\n");
            }
            return createFromDataString(stringBuilder.toString());
        } catch (FileNotFoundException exception) {
            return null;
        } catch (IOException exception) {
            return null;
        }
    }


    // -------------------------------------------------------------------------


    /**
     * Returns the reference to the set of nonterminal symbols.
     * @return Set of nonterminal symbols.
     */
    public Set<String> getNonterminalSymbols() {
        return this.nonterminalSymbols;
    }

    /**
     * Returns the reference to the set of terminal symbols.
     * @return Set of terminal symbols.
     */
    public Set<String> getTerminalSymbols() {
        return this.terminalSymbols;
    }

    /**
     * Returns the reference to the set of starting symbols.
     * @return Set of starting nonterminal symbols.
     */
    public Set<String> getStartingSymbols() {
        return this.startingSymbols;
    }

    
    /**
     * Returns the reference to the internal container with production rules.
     * @return Set of all production rules.
     */
    public Set<Rule> getProductionRules() {
        return this.productionRules;
    }


    // -------------------------------------------------------------------------


    /**
     * Returns the reference to the starting context.
     * @return Starting grammar context.
     */
    public Context getStartingContext() {
        return this.startingContext;
    }

    /**
     * Returns the reference to the empty context.
     * @return Empty grammar context.
     */
    public Context getEmptyContext() {
        return this.emptyContext;
    }

    /**
     * Returns the general context for a given nonterminal and terminal.
     * @param nonterminalSymbol Nonterminal symbol.
     * @param terminalSymbol Terminal symbol.
     * @return General grammar context reference.
     */
    public Context getGeneralContext(
        String nonterminalSymbol,
        String terminalSymbol
    ) {
        return this.grammarContexts.get(nonterminalSymbol).get(terminalSymbol);
    }


    // -------------------------------------------------------------------------


    /**
     * Serializes the entire grammar into the standard definition string.
     * @return Grammar definition data string.
     * @see #createFromDataString(java.lang.String)
     */
    public String serializeGrammar() {
        StringBuilder builder = new StringBuilder();
        for (String starting : this.startingSymbols) {
            builder.append(starting);
            builder.append(" ");
        }
        builder.append("\n");
        for (Rule rule : this.productionRules) {
            builder.append(rule.getNonterminalSymbol());
            builder.append(" ");
            builder.append(rule.getTerminalSymbol());
            builder.append(" ");
            builder.append(rule.getContentModel());
            builder.append("\n");
        }
        return builder.toString();
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Prints the debugging dump of the tree grammar.
     * @param indentPrefix Indentation string.
     */
    public void printDebugDump(String indentPrefix) {
        
        System.out.print(indentPrefix + "- nonterminal symbols:");
        for (String symbol : this.nonterminalSymbols) {
            System.out.print(" " + symbol);
        }
        System.out.println();
        
        System.out.print(indentPrefix + "- terminal symbols:");
        for (String symbol : this.terminalSymbols) {
            System.out.print(" " + symbol);
        }
        System.out.println();
        
        System.out.print(indentPrefix + "- starting nonterminals:");
        for (String symbol : this.startingSymbols) {
            System.out.print(" " + symbol);
        }
        System.out.println();
        
        System.out.println(indentPrefix + "- production rules:");
        for (Rule rule : this.productionRules) {
            System.out.println(
                indentPrefix + "  " +
                rule.getNonterminalSymbol() + " " +
                rule.getTerminalSymbol() + " " +
                rule.getContentModel()
            );
        }

        System.out.println(indentPrefix + "- grammar contexts:");
        for (String nonterminal : this.grammarContexts.keySet()) {
            Map<String, Context> map = this.grammarContexts.get(nonterminal);
            for (String terminal : map.keySet()) {
                System.out.println(
                    indentPrefix +
                    "  (n:" + nonterminal + "|t:" + terminal + "):"
                );
                map.get(terminal).printDebugDump(indentPrefix + "    ");
            }
        }

        System.out.println(indentPrefix + "- starting rule:");
        System.out.println(
            "   " +
            this.startingRule.getNonterminalSymbol() + " " +
            this.startingRule.getTerminalSymbol() + " " +
            this.startingRule.getContentModel()
        );

        System.out.println(indentPrefix + "- starting context:");
        this.startingContext.printDebugDump(indentPrefix + "    ");

        System.out.println(indentPrefix + "- empty rule:");
        System.out.println(
            "   " +
            this.emptyRule.getNonterminalSymbol() + " " +
            this.emptyRule.getTerminalSymbol() + " " +
            this.emptyRule.getContentModel()
        );

        System.out.println(indentPrefix + "- empty context:");
        this.emptyContext.printDebugDump(indentPrefix + "    ");

    }


    // -------------------------------------------------------------------------

}
