
package cz.cuni.mff.corrector.grammar;

import cz.cuni.mff.corrector.expression.Expression;
import java.util.Set;

/**
 * Regular tree grammar production rule representation.
 *
 * @author Martin Svoboda
 */
public class Rule {

    // -------------------------------------------------------------------------


    /**
     * Nonterminal on a right-hand side of the production rule.
     */
    private String nonterminalSymbol;

    /**
     * Terminal on a left-hand side of the production rule.
     */
    private String terminalSymbol;


    /**
     * Regular expression on a left-hand side of the production rule.
     */
    private String contentModel;

    /**
     * Set of nonterminal symbols occurring in the content model.
     */
    private Set<String> allowedNonterminals;

    /**
     * Constructed Glushkov automaton for an associated regular expression.
     */
    private Automaton glushkovAutomaton;


    // -------------------------------------------------------------------------


    /**
     * Creates a new instance of the production rule.
     * @param nonterminalSymbol Nonterminal on a right-hand side.
     * @param terminalSymbol Terminal on a left-hand side.
     * @param contentModel Regular expression on a left-hand side.
     */
    public Rule(
        String nonterminalSymbol,
        String terminalSymbol,
        String contentModel
    ) {

        // Storing internal rule fields
        this.nonterminalSymbol = nonterminalSymbol;
        this.terminalSymbol = terminalSymbol;

        // Parsing content model expression
        this.contentModel = contentModel;
        Expression basicModel = Expression.createExpression(contentModel);
        this.allowedNonterminals = basicModel.getSymbols();

        // Constructing a Glushkov automaton
        Expression markedModel = Expression.markExpression(basicModel);
        this.glushkovAutomaton = new Automaton(markedModel);
        
    }


    // -------------------------------------------------------------------------


    /**
     * Returns a nonterminal symbol from the production rule.
     * @return Nonterminal on a right-hand side.
     */
    public String getNonterminalSymbol() {
        return this.nonterminalSymbol;
    }

    /**
     * Returns a terminal symbol from the production rule.
     * @return Terminal on a right-hand side.
     */
    public String getTerminalSymbol() {
        return this.terminalSymbol;
    }


    /**
     * Returns a regular expression from the production rule.
     * @return Regular expression on a left-hand side.
     */
    public String getContentModel() {
        return this.contentModel;
    }

    /**
     * Returns the set of symbols with occurrence in the content model.
     * @return Set of nonterminals in a regular expression.
     */
    public Set<String> getAllowedNonterminals() {
        return this.allowedNonterminals;
    }

    /**
     * Returns the Glushkov automaton for an associated regular expression.
     * @return Associated Glushkov automaton.
     */
    public Automaton getGlushkovAutomaton() {
        return this.glushkovAutomaton;
    }


    // -------------------------------------------------------------------------

}
