
package cz.cuni.mff.corrector.graph;

import cz.cuni.mff.corrector.instruction.Instruction;
import cz.cuni.mff.corrector.intent.Intent;
import cz.cuni.mff.corrector.repair.NodeRepair;
import java.util.*;

/**
 * Implementation of the repairing multigraph.
 *
 * @author Martin Svoboda
 */
public class Graph {

    // -------------------------------------------------------------------------
    
    
    /**
     * Distinguished constant for the stratum zero.
     */
    public static int STRATUM_ZERO = 0;
    
    
    // -------------------------------------------------------------------------


    /**
     * Set of all vertices in the multigraph.
     */
    private Set<Vertex> graphVertices;

    /**
     * Set of all edges in the multigraph.
     */
    private Set<Edge> graphEdges;
    
    
    /**
     * Mapping from stratum number to automaton state and finally to vertex.
     */
    private Map<Integer, Map<String, Vertex>> strataClusters;


    /**
     * Mapping from vertices to sets of their ingoing edges.
     */
    private Map<Vertex, Set<Edge>> ingoingEdges;

    /**
     * Mapping from vertices to sets of their outgoing edges.
     */
    private Map<Vertex, Set<Edge>> outgoingEdges;



    // -------------------------------------------------------------------------


    /**
     * Set of all defined target states in the last stratum.
     */
    private Set<String> targetStates;

    /**
     * Number of strata in the completely constructed multigraph.
     */
    private int strataNumber;


    // -------------------------------------------------------------------------


    /**
     * Partially evaluated pathPrev function.
     */
    private Map<Vertex, Set<Edge>> pathPrev;

    /**
     * Partially evaluated pathCost function.
     */
    private Map<Vertex, Integer> pathCost;
    
    
    /**
     * Terminating reached cost of shortest paths to any target vertex.
     */
    private int termCost;


    // -------------------------------------------------------------------------


    /**
     * Creates a new and empty instance of the correction multigraph.
     * @param targetStates Set of all target states.
     * @param strataNumber Total number of strata.
     */
    public Graph(Set<String> targetStates, int strataNumber) {
        this.graphVertices = new HashSet<Vertex>();
        this.graphEdges = new HashSet<Edge>();
        this.strataClusters = new HashMap<Integer, Map<String, Vertex>>();
        this.ingoingEdges = new HashMap<Vertex, Set<Edge>>();
        this.outgoingEdges = new HashMap<Vertex, Set<Edge>>();
        this.targetStates = targetStates;
        this.strataNumber = strataNumber;
        this.pathPrev = new HashMap<Vertex, Set<Edge>>();
        this.pathCost = new HashMap<Vertex, Integer>();
        this.termCost = 0;
    }


    // -------------------------------------------------------------------------


    /**
     * Returns the set of all vertices in the multigraph.
     * @return Reference to the internal set of vertices.
     */
    public Set<Vertex> getGraphVertices() {
        return this.graphVertices;
    }

    /**
     * Returns the set of all edges in the multigraph.
     * @return Reference to the internal set of edges.
     */
    public Set<Edge> getGraphEdges() {
        return this.graphEdges;
    }


    // -------------------------------------------------------------------------


    /**
     * Returns the overall count of vertices in the multigraph.
     * @return Multigraph vertices count.
     */
    public int getVerticesCount() {
        return this.graphVertices.size();
    }

    /**
     * Returns the overall count of edges in the multigraph.
     * @return Multigraph edges count.
     */
    public int getEdgesCount() {
        return this.graphEdges.size();
    }


    // -------------------------------------------------------------------------


    /**
     * Adds a new vertex into the multigraph.
     * @param newVertex New vertex to be inserted.
     */
    public void addNewVertex(Vertex newVertex) {
        // Adding vertex into the base container
        this.graphVertices.add(newVertex);
        // Adding vertex into the stratum cluster
        Integer stratumNumber = newVertex.getStratumNumber();
        String automatonState = newVertex.getAutomatonState();
        // Accessing the stratum cluster
        Map<String, Vertex> stratumCluster =
            this.strataClusters.get(stratumNumber);
        if (stratumCluster == null) {
            stratumCluster = new HashMap<String, Vertex>();
            this.strataClusters.put(stratumNumber, stratumCluster);
        }
        // Adding vertex into the stratum cluster
        stratumCluster.put(automatonState, newVertex);
    }
    
    
    /**
     * Attempts to fetch the existing vertex from the multigraph.
     * @param stratumNumber Specified stratum number.
     * @param automatonState Specified automaton state name.
     * @return Existing vertex or <code>null</code>.
     */
    public Vertex getVertexItem(
        Integer stratumNumber,
        String automatonState
    ) {
        if (this.strataClusters.containsKey(stratumNumber)) {
            Map<String, Vertex> stratumCluster =
                this.strataClusters.get(stratumNumber);
            return stratumCluster.get(automatonState);
        } else {
            return null;
        }
    }
    
    /**
     * Detects whether the given vertex is already in the multigraph.
     * @param inspectedVertex Specified stratum number and automaton state.
     * @return Returns <code>true</code> if the given vertex already exists.
     */
    public boolean hasVertexItem(Vertex inspectedVertex) {
        int stratumNumber = inspectedVertex.getStratumNumber();
        String automatonState = inspectedVertex.getAutomatonState();
        if (this.strataClusters.containsKey(stratumNumber)) {
            Map<String, Vertex> stratumCluster =
                this.strataClusters.get(stratumNumber);
            return (stratumCluster.get(automatonState) != null);
        } else {
            return false;
        }
    }


    /**
     * Attempts to fetch the existing vertex from the last stratum.
     * @param automatonState Specified automaton state name.
     * @return Existing vertex or <code>null</code>.
     */
    public Vertex getTargetVertex(
        String automatonState
    ) {
        return getVertexItem(this.strataNumber, automatonState);
    }


    // -------------------------------------------------------------------------


    /**
     * Adds a new edge into the multigraph.
     * @param newEdge New edge to be inserted.
     */
    public void addNewEdge(Edge newEdge) {
        // Adding a new edge into the graph
        this.graphEdges.add(newEdge);
        // Updating the sets of ingoing edges
        Set<Edge> inEdges = this.ingoingEdges.get(newEdge.getTargetVertex());
        if (inEdges == null) {
            inEdges = new HashSet<Edge>(1);
            this.ingoingEdges.put(newEdge.getTargetVertex(), inEdges);
        }
        inEdges.add(newEdge);
        // Updating the sets of outgoing edges
        Set<Edge> outEdges = this.outgoingEdges.get(newEdge.getSourceVertex());
        if (outEdges == null) {
            outEdges = new HashSet<Edge>(1);
            this.outgoingEdges.put(newEdge.getSourceVertex(), outEdges);
        }
        outEdges.add(newEdge);
    }


    /**
     * Returns the set of all edges ingoing to the specified vertex.
     * @param baseVertex Specifies base vertex.
     * @return Internal set of ingoing edges or <code>null</code>.
     */
    public Set<Edge> getIngoingEdges(Vertex baseVertex) {
        return this.ingoingEdges.get(baseVertex);
    }

    /**
     * Returns the set of all edges outgoing from the specified vertex.
     * @param baseVertex Specifies base vertex.
     * @return Internal set of outgoing edges or <code>null</code>.
     */
    public Set<Edge> getOutgoingEdges(Vertex baseVertex) {
        return this.outgoingEdges.get(baseVertex);
    }


    // -------------------------------------------------------------------------


    /**
     * Returns the target states of the graph.
     * @return Reference to the internal set with target states.
     */
    protected Set<String> getTargetStates() {
        return this.targetStates;
    }


    // -------------------------------------------------------------------------


    /**
     * Returns the set of preceding edges on shortest paths.
     * @param graphVertex Specified base vertex.
     * @return Reference to the internal set or <code>null</code>.
     */
    public Set<Edge> getTracingEdges(Vertex graphVertex) {
        return this.pathPrev.get(graphVertex);
    }

    /**
     * Sets the new set of preceding edges on shortest paths.
     * @param graphVertex Specified base vertex.
     * @param prevEdges Reference to the new set of preceding edges.
     */
    public void setTracingEdges(Vertex graphVertex, Set<Edge> prevEdges) {
        this.pathPrev.put(graphVertex, prevEdges);
    }

    /**
     * Completely removes the set of preceding edges.
     * @param graphVertex Specified base vertex.
     */
    public void removeTracingEdges(Vertex graphVertex) {
        this.pathPrev.remove(graphVertex);
    }

    /**
     * Clears the existing set of preceding edges.
     * @param graphVertex Specified base vertex.
     */
    public void clearTracingEdges(Vertex graphVertex) {
        Set<Edge> prevSet = this.pathPrev.get(graphVertex);
        if (prevSet != null) {
            prevSet.clear();
        }
    }

    /**
     * Creates a new empty set of preceding edges and replaces the original.
     * @param graphVertex Specified base vertex.
     */
    public void initTracingEdges(Vertex graphVertex) {
        this.pathPrev.put(graphVertex, new HashSet<Edge>());
    }

    /**
     * Adds a new preceding edge into the set of a given base vertex.
     * Attempting to add an already stored edge do not cause duplicates.
     * @param graphVertex Specified base vertex.
     * @param precedingEdge Preceding edge to be added.
     */
    public void addTracingEdge(Vertex graphVertex, Edge precedingEdge) {
        Set<Edge> prevSet = this.pathPrev.get(graphVertex);
        if (prevSet == null) {
            prevSet = new HashSet<Edge>(1);
            this.pathPrev.put(graphVertex, prevSet);
        }
        prevSet.add(precedingEdge);
    }


    /**
     * Returns the cost of shortest paths leading to the given vertex.
     * If the cost is not defined, then <code>null</code> is returned.
     * @param graphVertex Specified base vertex.
     * @return Overall cost to reaching given vertex.
     */
    public Integer getTracingCost(Vertex graphVertex) {
        return this.pathCost.get(graphVertex);
    }

    /**
     * Sets a new path cost value for a given vertex.
     * @param graphVertex Specified base vertex.
     * @param newCost New value of the pathCost function.
     */
    public void setTracingCost(Vertex graphVertex, Integer newCost) {
        this.pathCost.put(graphVertex, newCost);
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the stored tracing cost for the target vertex of the edge.
     * @param inspectedEdge Reference to the edge to be inspected.
     * @return Stored value of the tracing cost or <code>null</code>.
     */
    public Integer getTracedCost(Edge inspectedEdge) {
        return getTracingCost(inspectedEdge.getTargetVertex());
    }
    
    /**
     * Returns the computed cost to the target vertex using this edge only.
     * @param inspectedEdge Reference to the edge to be inspected.
     * @return Computed value of the tracing cost or <code>null</code>.
     */
    public Integer getResolvedCost(Edge inspectedEdge) {
        return getTracingCost(inspectedEdge.getSourceVertex()) +
            inspectedEdge.getEdgeCost();
    }
    
    
    // -------------------------------------------------------------------------


    /**
     * Returns the terminating cost of the aggregated multigraph.
     * @return Minimal cost of shortest paths to any target vertex.
     */
    public int getTermCost() {
        return this.termCost;
    }


    /**
     * Updates the value of the terminating cost.
     * @param newCost New value of termCost function.
     */
    public void setTermCost(int newCost) {
        this.termCost = newCost;
    }


    // -------------------------------------------------------------------------


    /**
     * Generates all shortest paths ending in the specified target vertex.
     * @param outputSet Prepared output container for found paths.
     * @param targetVertex Target vertex of specified paths.
     */
    protected void generateShortestPaths(
        List<List<Edge>> outputSet,
        Vertex targetVertex
    ) {
        exploreShortestPaths(outputSet, new ArrayList<Edge>(), targetVertex);
    }


    /**
     * Recursive function for reverse traversal of shortest paths.
     * Finds all shortest paths from the source vertex to the specified local
     * target vertex and returns these paths in the prepared container.
     * @param outputSet Prepared output container for found paths.
     * @param partialPath Partially reconstructed path.
     * @param currentVertex Current ending vertex to explore from.
     */
    private void exploreShortestPaths(
        List<List<Edge>> outputSet,
        List<Edge> partialPath,
        Vertex currentVertex
    ) {

        // Fetching the set of precesing edges
        Set<Edge> prevEdges = this.getTracingEdges(currentVertex);

        // Terminating the constructed path in the source vertex
        if ((prevEdges == null) || (prevEdges.isEmpty())) {

            List<Edge> newPath = new ArrayList<Edge>(partialPath);
            outputSet.add(newPath);

        // Exploring the next preceding directins towards the source
        } else {

            partialPath.add(0, null);
            for (Edge prevEdge : prevEdges) {
                partialPath.set(0, prevEdge);
                exploreShortestPaths(
                    outputSet,
                    partialPath,
                    prevEdge.getSourceVertex()
                );
            }
            partialPath.remove(0);

        }

    }
    
    
    /**
     * Generates the set of all shortest paths in the aggregated multigraph.
     * @return Set of paths with minimal cost to any target vertex.
     */
    public List<List<Edge>> fetchShortestPaths() {
        List<List<Edge>> pathsSet = new ArrayList<List<Edge>>();
        for (String targetState : this.getTargetStates()) {
            Vertex targetVertex = this.getTargetVertex(targetState);
            if (
                (targetVertex == null) ||
                (this.getTracingCost(targetVertex) == null)
            ) {
                continue;
            }
            if (this.getTracingCost(targetVertex) == this.getTermCost()) {
                generateShortestPaths(pathsSet, targetVertex);
            }
        }
        return pathsSet;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Comparator for sorting (shortest) paths.
     */
    public class PathsComparator implements Comparator<List<Edge>> {

        public int compare(List<Edge> p1, List<Edge> p2) {
            
            // Fetching path sizes
            int s1 = p1.size();
            int s2 = p2.size();
            
            // Comparing edge by edge
            int i = 0;
            while ((i < s1) && (i < s2)) {
                Edge e1 = p1.get(i);
                Edge e2 = p2.get(i);
                int c = e1.compareTo(e2);
                if (c != 0) {
                    return c;
                }
                i++;
            }
            
            // Comparing path sizes
            return Integer.compare(s1, s2);
            
        }
        
    }
    
    
    /**
     * Automatically created comparator for paths in this multigraph.
     */
    public final PathsComparator pathsComparator = new PathsComparator();
    
    
    // -------------------------------------------------------------------------


    /**
     * Comparator ordering vertices using pathCost function.
     * As less are treated those vertices that have defined pathCost, have
     * lower pathCost or are located at a higher stratum.
     */
    public class ReachedVerticesComparator implements Comparator<Vertex> {

        public int compare(Vertex o1, Vertex o2) {
            Integer c1 = getTracingCost(o1);
            Integer c2 = getTracingCost(o2);
            return compareNulls(o1, o2, c1, c2);
        }

        /**
         * Partial comparator giving preference to defined costs.
         * @param o1 Reference to the first vertex.
         * @param o2 Reference to the second vertex.
         * @param c1 Cost of the first vertex.
         * @param c2 Cost of the second vertex.
         * @return Comparison value.
         */
        private int compareNulls(Vertex o1, Vertex o2, Integer c1, Integer c2) {
            if ((c1 == null) && (c2 == null)) {
                return compareStrata(o1, o2);
            } else if (c1 == null) {
                return 1;
            } else if (c2 == null) {
                return -1;
            } else {
                return compareCosts(o1, o2, c1, c2);
            }
        }

        /**
         * Partial comparator giving preference to lower costs.
         * @param o1 Reference to the first vertex.
         * @param o2 Reference to the second vertex.
         * @param c1 Cost of the first vertex.
         * @param c2 Cost of the second vertex.
         * @return Comparison value.
         */
        private int compareCosts(Vertex o1, Vertex o2, Integer c1, Integer c2) {
            if (c1 < c2) {
                return -1;
            } else if (c1 > c2) {
                return 1;
            } else {
                return compareStrata(o1, o2);
            }
        }

        /**
         * Partial comparator giving preference to higher strata.
         * @param o1 Reference to the first vertex.
         * @param o2 Reference to the second vertex.
         * @return Comparison value.
         */
        private int compareStrata(Vertex o1, Vertex o2) {
            if (o1.getStratumNumber() > o2.getStratumNumber()) {
                return -1;
            } else if (o1.getStratumNumber() < o2.getStratumNumber()) {
                return 1;
            } else {
                return o1.getAutomatonState().compareTo(o2.getAutomatonState());
            }
        }

    }


    /**
     * Initialized internal comparator for reached vertices.
     */
    private ReachedVerticesComparator comparatorReached =
        new ReachedVerticesComparator();


    /**
     * Returns the reference to the initialized vertices comparator.
     * @return Reached vertices comparator instance.
     */
    public ReachedVerticesComparator getComparatorReached() {
        return this.comparatorReached;
    }


    // -------------------------------------------------------------------------


    /**
     * Prints the debugging dump of the repairing multigraph.
     * @param indentPrefix Indentation string.
     */
    public void printDebugDump(String indentPrefix) {

        System.out.print(indentPrefix + "- graph vertices:");
        for (Vertex vertex : this.graphVertices) {
            System.out.print(
                " (" + vertex.getStratumNumber() +
                "|" + vertex.getAutomatonState() + ")"
            );
        }
        System.out.println();

        System.out.println(indentPrefix + "- graph edges:");
        for (Edge edge : this.graphEdges) {
            Intent intent = edge.getCorrectionIntent();
            System.out.print(
                indentPrefix +
                "  - [(" + edge.getSourceStratum() +
                "|" + edge.getSourceState() + ")-" +
                "(" + edge.getTargetStratum() +
                "|" + edge.getTargetState() + ")]" +
                " cost=" + ((edge.getIntentRepair() != null) ? edge.getEdgeCost() : "?") +
                " type=" + ((intent == null) ? "?" : intent.getIntentType()) +
                " instr:"
            );
            if (intent != null) {
                NodeRepair repair = intent.getNodeRepair();
                for (Instruction i : repair.getInstructionsList()) {
                    System.out.print(" " + i.getTitleString());
                }
            } else {
                System.out.print("?");
            }
            System.out.println();
        }

        System.out.print(indentPrefix + "- target states:");
        for (String state : this.targetStates) {
            System.out.print(" " + state);
        }
        System.out.println();

        System.out.println(
            indentPrefix + "- strata count: " + this.strataNumber
        );

        System.out.println(indentPrefix + "- path prevs:");
        for (Vertex vertex : this.pathPrev.keySet()) {
            System.out.print(
                indentPrefix +
                "  - (" + vertex.getStratumNumber() +
                "|" + vertex.getAutomatonState() + "):"
            );
            for (Edge prev : this.pathPrev.get(vertex)) {
                System.out.print(
                    " (" + prev.getSourceStratum() +
                    "|" + prev.getSourceState() + ")"
                );
            }
            System.out.println();
        }

        System.out.println(indentPrefix + "- path costs:");
        for (Vertex vertex : this.pathCost.keySet()) {
            System.out.println(
                indentPrefix +
                "  - (" + vertex.getStratumNumber() +
                "|" + vertex.getAutomatonState() + "): " +
                this.pathCost.get(vertex)
            );
        }

        System.out.println(indentPrefix + "- strata clusters:");
        for (Integer stratum: this.strataClusters.keySet()) {
            System.out.print(indentPrefix + "  - " + stratum + ":");
            Map<String, Vertex> map = this.strataClusters.get(stratum);
            for (String state : map.keySet()) {
                System.out.print(
                    " (" + map.get(state).getStratumNumber() +
                    "|" + map.get(state).getAutomatonState() + ")"
                );
            }
            System.out.println();
        }

        System.out.println(indentPrefix + "- ingoing edges:");
        for (Vertex vertex: this.ingoingEdges.keySet()) {
            System.out.print(
                indentPrefix +
                "  - (" + vertex.getStratumNumber() +
                "|" + vertex.getAutomatonState() + "):"
            );
            Set<Edge> edges = this.ingoingEdges.get(vertex);
            for (Edge edge : edges) {
                System.out.print(
                    " [(" + edge.getSourceStratum() +
                    "|" + edge.getSourceState() + ")-" +
                    "(" + edge.getTargetStratum() +
                    "|" + edge.getTargetState() + ")]"
                );
            }
            System.out.println();
        }

        System.out.println(indentPrefix + "- outgoing edges:");
        for (Vertex vertex: this.outgoingEdges.keySet()) {
            System.out.print(
                indentPrefix +
                "  - (" + vertex.getStratumNumber() +
                "|" + vertex.getAutomatonState() + "):"
            );
            Set<Edge> edges = this.outgoingEdges.get(vertex);
            for (Edge edge : edges) {
                System.out.print(
                    " [(" + edge.getSourceStratum() +
                    "|" + edge.getSourceState() + ")-" +
                    "(" + edge.getTargetStratum() +
                    "|" + edge.getTargetState() + ")]"
                );
            }
            System.out.println();
        }
        
    }


    // -------------------------------------------------------------------------

}
