
package cz.cuni.mff.corrector.graph;

/**
 * Vertex of the repairing multigraph.
 * 
 * Vertex instances can be used in sets as their identification is properly
 * resolved using immutable stratum number and automaton state attributes. 
 *
 * @author Martin Svoboda
 */
public class Vertex implements Comparable<Vertex> {

    // -------------------------------------------------------------------------


    /**
     * Number of the stratum in the multigraph.
     */
    private final int stratumNumber;

    /**
     * Name of the automaton state.
     */
    private final String automatonState;


    // -------------------------------------------------------------------------


    /**
     * Creates a new vertex instance.
     * @param stratumNumber Number of the stratum in the multigraph.
     * @param automatonState Name of the automaton state.
     */
    public Vertex(int stratumNumber, String automatonState) {
        this.stratumNumber = stratumNumber;
        this.automatonState = automatonState;
    }


    // -------------------------------------------------------------------------


    /**
     * Returns the stratum number.
     * @return Number of the stratum in the multigraph.
     */
    public int getStratumNumber() {
        return this.stratumNumber;
    }

    /**
     * Returns the automaton state.
     * @return Name of the automaton state.
     */
    public String getAutomatonState() {
        return this.automatonState;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the vertex description string.
     * @return Human readable vertex serialization.
     */
    public String getSerializedString() {
        return "(" + this.stratumNumber + "|" + this.automatonState + ")";
    }


    // -------------------------------------------------------------------------

    
    @Override
    public boolean equals(Object obj) {
        if (obj instanceof Vertex) {
            Vertex v = (Vertex)obj;
            return (
                (this.stratumNumber == v.stratumNumber) &&
                (this.automatonState.equals(v.automatonState))
            );
        } else {
            return false;
        }
    }
    

    @Override
    public int hashCode() {
        int hash = 5;
        hash = 73 * hash + this.stratumNumber;
        hash = 73 * hash + this.automatonState.hashCode();
        return hash;
    }
    
    
    public int compareTo(Vertex v) {
        // Comparing strate numbers
        int cStrata = Integer.compare(this.stratumNumber, v.stratumNumber);
        if (cStrata != 0) {
            return cStrata;
        }
        // Comparing automaton states
        return this.automatonState.compareTo(v.automatonState);
    }
    

    // -------------------------------------------------------------------------

}
