
package cz.cuni.mff.corrector.intent;

import cz.cuni.mff.corrector.grammar.Automaton;
import cz.cuni.mff.corrector.grammar.Context;
import cz.cuni.mff.corrector.grammar.Grammar;
import cz.cuni.mff.corrector.grammar.Rule;
import cz.cuni.mff.corrector.graph.Vertex;
import cz.cuni.mff.corrector.instruction.Instruction;
import cz.cuni.mff.corrector.repair.NodeRepair;
import cz.cuni.mff.corrector.tree.Node;
import java.util.*;

/**
 * Abstract ancestor for implemented correction intents.
 * 
 * Intent objects can also be automatically tested for equality, however,
 * only correction intents within only one particular multigraph can be
 * compared to each other! Note that this condition cannot be tested and
 * when breached the returned results make not sense!
 *
 * @author Martin Svoboda
 */
public abstract class Intent {
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Enumeration of all implemented correction intent types.
     */
    public enum Type {
        
        /**
         * Correction intent type: CORRECT.
         */
        CORRECT("correct", "COR", "C"),
        /**
         * Correction intent type: INSERT.
         */
        INSERT("insert", "INS", "I"),
        /**
         * Correction intent type: DELETE.
         */
        DELETE("delete", "DEL", "D"),
        /**
         * Correction intent type: REPAIR.
         */
        REPAIR("repair", "REP", "P"),
        /**
         * Correction intent type: RENAME.
         */
        RENAME("rename", "REN", "N");
        
        /**
         * Full name (e.g. "insert") of the intent type.
         */
        public final String fullName;
        /**
         * Full code (e.g. "INS") of the intent type.
         */
        public final String fullCode;
        /**
         * Tiny code (e.g. "I") of the intent type.
         */
        public final String tinyCode;
        
        /**
         * Creates a new intent type enumeration constant instance.
         * @param fullName Full name of the intent type.
         * @param fullCode Full code of the intent type.
         * @param tinyCode Tiny code of the intent type.
         */
        Type(String fullName, String fullCode, String tinyCode) {
            this.fullName = fullName;
            this.fullCode = fullCode;
            this.tinyCode = tinyCode;
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    // Intent action attributes
    
    
    /**
     * Optional reference to the base node associated with the node repair.
     * In case of correct or insert intents <code>null</code> reference.
     */
    protected Node baseNode;
    
    /**
     * Optional associated node repair (repairing instruction).
     */
    protected NodeRepair nodeRepair;
    
    /**
     * Pair of the initial stratum number and automaton state.
     */
    protected Vertex initialVertex;
    
    /**
     * Pair of the ending stratum number and automaton state.
     */
    protected Vertex endingVertex;
    
    
    // -------------------------------------------------------------------------
    // Intent assignment attributes
    
    
    /**
     * Sequence of nodes to be processed.
     */
    protected List<Node> nodeSequence;
    
    /**
     * Reference to the nested grammar context.
     */
    protected Context grammarContext;
    
    /**
     * Optional chain of nested grammar contexts.
     * This chain is only used in case of INSERT correction intents.
     */
    protected List<Context> contextChain;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Auxiliary item for storing complete intent signature.
     */
    private String intentSignature;
    
    /**
     * Separator string used for imploding signature components.
     */
    public static String SEPARATOR = "-";
    
    
    /**
     * Auxiliary item for storing complete intent identifier.
     * This identifier enables distinguishing between particular correction
     * intents, but only within one given parental intent and its multigraph.
     */
    private String intentIdentifier;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates a new instance of the correction intent.
     * @param baseNode Optional associated base node reference.
     * @param nodeRepair Optional repairing instruction.
     * @param initialVertex Pair of the initial stratum and state.
     * @param endingVertex Pair of the ending stratum and state.
     * @param nodeSequence Sequence of nested nodes to be processed.
     * @param grammarContext Nested grammar context reference.
     * @param contextChain Optional chain of grammar contexts.
     */
    protected Intent(
        Node baseNode,
        NodeRepair nodeRepair,
        Vertex initialVertex,
        Vertex endingVertex,
        List<Node> nodeSequence,
        Context grammarContext,
        List<Context> contextChain
    ) {
        
        // Storing action components
        this.baseNode = baseNode;
        this.nodeRepair = nodeRepair;
        this.initialVertex = initialVertex;
        this.endingVertex = endingVertex;
        
        // Storing assignment components
        this.nodeSequence = nodeSequence;
        this.grammarContext = grammarContext;
        this.contextChain = contextChain;
        
        // Initialising auxiliary data
        this.intentSignature = computeIntentSignature();
        this.intentIdentifier = computeIntentIdentifier();
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the type of the given correction intent.
     * @return Correction intent type enumeration constant.
     */
    public abstract Type getIntentType();
    
    
    // -------------------------------------------------------------------------
    

    /**
     * Returns the node repair associated with the intent.
     * @return Associated node repair structure reference.
     */
    public NodeRepair getNodeRepair() {
        return this.nodeRepair;
    }
    
    /**
     * Returns the optional reference to the base node.
     * @return Base node from the original data tree.
     */
    public Node getBaseNode() {
        return this.baseNode;
    }
    
    /**
     * Returns the pair of the initial stratum number and automaton state.
     * @return Initial vertex instance.
     */
    public Vertex getInitialVertex() {
        return this.initialVertex;
    }
    
    /**
     * Returns the pair of the ending stratum number and automaton state.
     * @return Ending vertex instance.
     */
    public Vertex getEndingVertex() {
        return this.endingVertex;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the sequence of nested nodes to be processed.
     * @return Reference to the internal list of nodes.
     */
    public List<Node> getNodesSequence() {
        return this.nodeSequence;
    }
    
    /**
     * Returns the number of nodes in the sequence to be processed.
     * @return Length of the nested node sequence.
     */
    public int getSequenceLength() {
        return this.nodeSequence.size();
    }
    
    /**
     * Returns the reference to the nested grammar context.
     * @return Grammar context reference.
     */
    public Context getGrammarContext() {
        return this.grammarContext;
    }
    
    /**
     * Returns the optional associated chain of contexts.
     * @return Reference to the internal context chain container.
     */
    public List<Context> getContextChain() {
        return this.contextChain;
    }
    
    /**
     * Returns the set of correction intent types allowed for nesting.
     * @return Set of intent types allowed for nested intents.
     */
    public abstract Set<Type> getAllowedIntents();
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the source state.
     * This state corresponds to the initial state of the given automaton
     * and corresponding vertex at the zeroth stratum of the multigraph.
     * @return Name of the source state from the zeroth stratum.
     */
    public String getSourceState() {
        return this.grammarContext.getGlushkovAutomaton().getInitialState();
    }
    
    /**
     * Returns the set of target states.
     * This set corresponds to all accepting states of the given automaton
     * and corresponding vertices at the last stratum of the multigraph.
     * @return Name of the source state from the zeroth stratum.
     */
    public Set<String> getTargetStates() {
        return this.grammarContext.getGlushkovAutomaton().getAcceptingStates();
    }
    
    
    /**
     * Returns the number of the last stratum for the nested multigraph.
     * Note that this function only resolves the number of strata the given
     * correction multigraph should have, i.e. the number of nodes in the
     * associated sequence to be processed.
     * @return Number of the last stratum in the resulting multigraph.
     */
    public int getLastStratum() {
        return this.nodeSequence.size();
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Generates the set of all permitted nested correction intents.
     * @param parentIntent Reference to the parental intent.
     * @param initialStratum Number of the initial stratum.
     * @param initialState Name of the initial automaton state.
     * @param treeGrammar Reference to the tree grammar.
     * @param enabledIntents Configuration set of enabled intent types.
     * @return List of generated nested intents.
     */
    public static List<Intent> generateNestedIntents(
        Intent parentIntent,
        int initialStratum,
        String initialState,
        Grammar treeGrammar,
        EnumSet<Intent.Type> enabledIntents
    ) {
        
        // Preparing an output intents set
        List<Intent> outputList = new ArrayList<Intent>();

        // Fetching allowed transitions
        Automaton parentAutomaton =
            parentIntent.getGrammarContext().getGlushkovAutomaton();
        Map<String, String> transitions =
            parentAutomaton.getTransitionRules(initialState);
        
        // Generating all nested intents
        if (enabledIntents.contains(Intent.Type.INSERT)) {
            IntentInsert.generateNestedIntents(
                outputList, parentIntent,
                initialStratum, initialState, treeGrammar, transitions
            );
        }
        if (enabledIntents.contains(Intent.Type.DELETE)) {
            IntentDelete.generateNestedIntents(
                outputList, parentIntent,
                initialStratum, initialState, treeGrammar
            );
        }
        if (enabledIntents.contains(Intent.Type.REPAIR)) {
            IntentRepair.generateNestedIntents(
                outputList, parentIntent,
                initialStratum, initialState, treeGrammar, transitions
            );
        }
        if (enabledIntents.contains(Intent.Type.RENAME)) {
            IntentRename.generateNestedIntents(
                outputList, parentIntent,
                initialStratum, initialState, treeGrammar, transitions
            );
        }
        
        // Returning generated intents
        return outputList;
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the complete intent signature string.
     * @return Intent signature string.
     */
    public String getGlobalSignature() {
        return this.intentSignature;
    }
    
    
    /**
     * Computes the intent signature string.
     * @return Fully composed global intent signature string.
     */
    protected abstract String computeIntentSignature();
    
    
    /**
     * Generates the signature string for the associated context chain.
     * The output is of the form: [-Ni-ti]*
     * @see #SEPARATOR
     * @return Serialized string with non/terminal symbol pairs.
     */
    public String computeChainSignature() {
        
        if ((this.contextChain == null) || (this.contextChain.isEmpty())) {
            return "";
        }
        
        StringBuilder sb = new StringBuilder();
        for (Context contextItem : this.contextChain) {
            sb.append(Intent.SEPARATOR);
            sb.append(contextItem.getBaseRule().getNonterminalSymbol());
            sb.append(Intent.SEPARATOR);
            sb.append(contextItem.getBaseRule().getTerminalSymbol());
        }
        return sb.toString();
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the locally valid intent identifier.
     * This identifier can be used within the parental correction multigraph
     * to correctly distinguish between all its nested correction intents.
     * Note that usage of intent signatures would not be sufficient for
     * this purpose!
     * @return Local intent identifier string.
     */
    public String getLocalIdentifier() {
        return this.intentIdentifier;
    }
    
    
    /**
     * Computes the intent local identifier string.
     * The prefix of this id is based on the description of action location,
     * i.e. is formed from the source and targets vertices. Next, the postfix
     * part is dependent on particular intent types.
     * @return Fully composed local intent identifier string.
     */
    protected String computeIntentIdentifier() {
        return computeIdentifierPrefix() + computeIdentifierPostfix();
    }
    
    /**
     * Computes the prefix part of the intent local identifier.
     * @return Prefix of the local identifier.
     */
    protected String computeIdentifierPrefix() {
        if ((this.initialVertex == null) || (this.endingVertex == null)) {
            return "";
        }
        return
            this.initialVertex.getStratumNumber() + SEPARATOR +
            this.initialVertex.getAutomatonState()+ SEPARATOR +
            this.endingVertex.getStratumNumber() + SEPARATOR +
            this.endingVertex.getAutomatonState() + SEPARATOR;
    }
    
    /**
     * Computes the postfix part of the intent local identifier.
     * @return Postfix of the local identifier.
     */
    protected abstract String computeIdentifierPostfix();
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Prints the debugging dump of the correction intent.
     * @param indentPrefix Indentation string.
     */
    public void printDebugDump(String indentPrefix) {
        
        System.out.println(
            indentPrefix + "- system identifier: " + this
        );
        
        System.out.println(
            indentPrefix + "- intent type: " + this.getIntentType()
        );
        
        if (this.getIntentType() != Intent.Type.CORRECT) {
            
            System.out.print(
                indentPrefix + "- node repair: " +
                this.getNodeRepair().getRepairCost() + " |"
            );
            for (Instruction i : this.getNodeRepair().getInstructionsList()) {
                System.out.print(" " + i.getTitleString());
            }
            System.out.println();
            
            if (this.getIntentType() == Intent.Type.INSERT) {
                IntentInsert ii = (IntentInsert)this;
                System.out.print(indentPrefix + "- insert chain:");
                for (Context ctx : ii.getContextChain()) {
                    System.out.print(
                        " (n:" + ctx.getBaseRule().getNonterminalSymbol() +
                        "|t:" + ctx.getBaseRule().getTerminalSymbol() + ")"
                    );
                }
                System.out.println();
            }
            
            System.out.println(
                indentPrefix + "- initial vertex: " +
                this.initialVertex.getSerializedString()
            );
            System.out.println(
                indentPrefix + "- ending vertex: " +
                this.endingVertex.getSerializedString()
            );
            
        }
        
        System.out.print(
            indentPrefix + "- node sequence: " +
            this.getSequenceLength()
        );
        for (Node node : this.getNodesSequence()) {
            System.out.print(
                " (l:" + node.getNodeLabel() +
                "|t:" + node.getNodeType() + ")"
            );
        }
        System.out.println();
        
        Rule rule = this.getGrammarContext().getBaseRule();
        System.out.println(
            indentPrefix + "- grammar context: " +
            rule.getNonterminalSymbol() + " " +
            rule.getTerminalSymbol() + " " +
            rule.getContentModel()
        );
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Note that this function can only be used when comparing correction
     * intents from within the same correction multigraph!
     */
    @Override
    public boolean equals(Object obj) {
        if (obj instanceof Intent) {
            Intent i = (Intent)obj;
            return (this.getLocalIdentifier().equals(i.getLocalIdentifier()));
        } else {
            return false;
        }
    }
    
    
    /**
     * Note that this function can only be used when comparing correction
     * intents from within the same correction multigraph!
     */
    @Override
    public int hashCode() {
        int hash = 3;
        hash = 17 * hash + this.getLocalIdentifier().hashCode();
        return hash;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Provider of unique identifiers for intent objects.
     */
    private static int ID_PROVIDER = 1;
    
    /**
     * Unique assigned identifier to each intent instance.
     */
    public final int ID = ID_PROVIDER++;
    
    
    /**
     * Returns the globally unique intent instance label.
     * This label can be used for debugging purposes to globally identify
     * and distinguish all created intent instances even across different
     * correction multigraphs.
     * The label is of the form: signature@id
     * @return Unique label of the intent instance.
     */
    public String getUniqueLabel() {
        return this.getGlobalSignature() + "@" + ID;
    }
    
    
    /**
     * Resets the current value of the intent identifier provider.
     */
    public static final void resetIdProvider() {
        ID_PROVIDER = 1;
    }
    
    
    // -------------------------------------------------------------------------
    
}
