
package cz.cuni.mff.corrector.intent;

import cz.cuni.mff.corrector.edit.Edit;
import cz.cuni.mff.corrector.grammar.Context;
import cz.cuni.mff.corrector.grammar.Grammar;
import cz.cuni.mff.corrector.graph.Vertex;
import cz.cuni.mff.corrector.instruction.InstructionRemoveLeaf;
import cz.cuni.mff.corrector.repair.Mend;
import cz.cuni.mff.corrector.repair.NodeRepairDeleteSubtree;
import cz.cuni.mff.corrector.tree.Node;
import java.util.*;

/**
 * Implementation of the delete correction intent.
 *
 * @author Martin Svoboda
 */
public class IntentDelete extends Intent {

    // -------------------------------------------------------------------------
    
    
    /**
     * List of all intent types allowed in further nesting.
     */
    private static final Set<Type> ALLOWED_INTENTS = EnumSet.of(Type.DELETE);
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates a new instance of the delete correction intent.
     * @param baseNode Optional associated base node reference.
     * @param initialVertex Pair of the initial stratum and state.
     * @param endingVertex Pair of the ending stratum and state.
     * @param nodeSequence Sequence of nested nodes to be processed.
     * @param grammarContext Nested grammar context reference.
     */
    public IntentDelete(
        Node baseNode,
        Vertex initialVertex,
        Vertex endingVertex,
        List<Node> nodeSequence,
        Context grammarContext
    ) {
        
        // Constructing core components
        super(
            baseNode,
            new NodeRepairDeleteSubtree(new InstructionRemoveLeaf()),
            initialVertex,
            endingVertex,
            nodeSequence,
            grammarContext,
            null
        );
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    @Override
    public Type getIntentType() {
        return Type.DELETE;
    }
    
    @Override
    public Set<Type> getAllowedIntents() {
        return ALLOWED_INTENTS;
    }
    
    @Override
    protected String computeIntentSignature() {
        return Type.DELETE.fullCode + Intent.SEPARATOR + getBaseNode();
    }
    
    @Override
    protected String computeIdentifierPostfix() {
        return Type.DELETE.fullCode;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Generates the set of all permitted nested delete intents.
     * @param outputSet Output container for generated intents.
     * @param parentIntent Parental correction intent reference.
     * @param initialStratum Number of the initial stratum.
     * @param initialState Name of the initial state.
     * @param treeGrammar Reference to the tree grammar.
     */
    public static void generateNestedIntents(
        List<Intent> outputSet,
        Intent parentIntent,
        int initialStratum,
        String initialState,
        Grammar treeGrammar
    ) {
        
        // Resolving necessary prerequisites
        boolean isNestingAllowed =
            (parentIntent.getAllowedIntents().contains(Intent.Type.DELETE));
        boolean isLastStratum =
            (initialStratum == parentIntent.getNodesSequence().size());
        
        // Further nesting is not possible
        if (!isNestingAllowed || isLastStratum) {
            return;
        }
        
        // Referencing the following node
        Node inspectedNode = parentIntent.nodeSequence.get(initialStratum);
        
        // Creating a nested correction intent
        IntentDelete newIntent = new IntentDelete(
            inspectedNode,
            new Vertex(initialStratum, initialState),
            new Vertex(initialStratum + 1, initialState),
            inspectedNode.getChildNodes(),
            treeGrammar.getEmptyContext()
        );
        
        // Adding the intent into the output set
        outputSet.add(newIntent);
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Generates all mends for the particular intent repair of delete type.
     * @param nodeEdit Sequence of edit operations repairing the node.
     * @param sequenceMends Set of all mends of the sequence repair.
     */
    public static void generateIntentFix(
        List<Edit> nodeEdit,
        List<Mend> sequenceMends
    ) {
        
        // Processing individual sequence mends
        for (Mend seqMend : sequenceMends) {
            
            // Adjusting edit operations from the sequence mend
            seqMend.applyRootPrepending(0);
            // Placing clones at the end of the original list
            Edit.cloneIdenticalSequenceAfter(
                nodeEdit, seqMend.getEditSequence()
            );
            
        }
        
    }
    
    
    // -------------------------------------------------------------------------

}
