
package cz.cuni.mff.corrector.tree;

import java.util.ArrayList;
import java.util.List;

/**
 * Data tree node representation
 *
 * @author Martin Svoboda
 */
public abstract class Node {
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Enumeration of node types.
     */
    public enum Type {
        /**
         * Node type: element node.
         */
        NODE_ELEMENT,
        /**
         * Node type: data node.
         */
        NODE_DATA;
    }


    // -------------------------------------------------------------------------


    /**
     * String token for a label of data nodes.
     */
    public static String TOKEN_DT_DATA = "#data";

    /**
     * Undefined value for data nodes.
     */
    public static String TOKEN_DV_UNDEFINED = null;


    // -------------------------------------------------------------------------

    
    /**
     * Opening parenthesis for an element starting tag.
     */
    public static String TOKEN_TG_OPEN_START = "<";

    /**
     * Closing parenthesis for an empty element tag.
     */
    public static String TOKEN_TG_CLOSE_EMPTY = "/>";

    /**
     * Opening parenthesis for an element ending tag.
     */
    public static String TOKEN_TG_OPEN_END = "</";

    /**
     * Closing parenthesis for a non-empty element tags.
     */
    public static String TOKEN_TG_CLOSE = ">";


    /**
     * End of line char used for serialization.
     */
    public static String TOKEN_SE_EOL = "\n";

    /**
     * String used for elements indentation during serialization.
     */
    public static String TOKEN_SE_INDENT = "    ";
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Provider of unique identifiers for node objects.
     */
    private static int idProvider = 0;
    
    
    /**
     * Assigned node identifier value.
     */
    private final int nodeId;


    // -------------------------------------------------------------------------


    /**
     * Initializes the node structure.
     */
    protected Node() {
        this.nodeId = idProvider++;
    }


    // -------------------------------------------------------------------------


    /**
     * Translates the given sequence of nodes to their labels.
     * @param listNodes Input sequence of nodes.
     * @return Sequence of labels.
     */
    public static List<String> getNodesLabels(List<Node> listNodes) {
        List<String> listLabels = new ArrayList<String>(listNodes.size());
        for (Node node : listNodes) {
            listLabels.add(node.getNodeLabel());
        }
        return listLabels;
    }


    // -------------------------------------------------------------------------


    /**
     * Returns the type of a given node.
     * @return Node type enumeration constant.
     */
    public abstract Type getNodeType();


    /**
     * Returns the node label.
     * @return Node label string.
     */
    public abstract String getNodeLabel();


    /**
     * Returns the list of all child nodes.
     * @return List of child nodes.
     */
    public abstract List<Node> getChildNodes();


    /**
     * Recursively counts all nodes in the data tree.
     * @return Overall number of nodes.
     */
    public abstract int countTreeNodes();
    
    
    /**
     * Computes the basic statistics about this subtree.
     * @return Container with computed statistics.
     */
    public abstract Stats getTreeStatistics();


    // -------------------------------------------------------------------------
    

    /**
     * Serializes the node into an XML string.
     * @param builder String builder for serialization output.
     */
    public abstract void serializeNode(StringBuilder builder);


    /**
     * Serializes the node into an indented XML string.
     * @param builder String builder for serialization output.
     * @param level Level of recursive nesting for indentation.
     */
    public abstract void serializeNode(StringBuilder builder, int level);


    /**
     * Generates indentation string used for indented serialization.
     * @param builder String builder for serialization output.
     * @param level Level of recursive nesting for indentation.
     */
    public static void generateIndent(StringBuilder builder, int level) {
        for (int i = 0; i < level; i++) {
            builder.append(Node.TOKEN_SE_INDENT);
        }
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the node identifier.
     * @return Assigned node identifier value.
     */
    public int getNodeIdentifier() {
        return this.nodeId;
    }
    
    
    @Override
    public String toString() {
        return String.valueOf(this.nodeId);
    }


    // -------------------------------------------------------------------------

}
