
package cz.cuni.mff.corrector.tree;

import java.util.ArrayList;
import java.util.List;

/**
 * Representation of the position address in a data tree.
 * 
 * @author Martin Svoboda
 */
public class Position {

    // -------------------------------------------------------------------------
    
    
    /**
     * Internal sequence of integers representing the address.
     */
    private List<Integer> basePosition;
    
    
    // -------------------------------------------------------------------------
    

    /**
     * Creates a new and empty position.
     */
    public Position() {
        this.basePosition = new ArrayList<Integer>();
    }

    /**
     * Creates a new position based on a single item address.
     * @param simplePosition Integer representing simple number of node.
     */
    public Position(int simplePosition) {
        this.basePosition = new ArrayList<Integer>();
        this.basePosition.add(simplePosition);
    }

    /**
     * Creates a new position from the prepared sequence of numbers.
     * Note that the provided sequence becomes the property of the position!
     * This means that the entire list and even all its items are adopted.
     * @param basePosition Sequence of simple positions to be seized.
     */
    public Position(List<Integer> basePosition) {
        this.basePosition = basePosition;
    }


    // -------------------------------------------------------------------------


    /**
     * Serializes the position into a readable string.
     * @return Serialized position string.
     */
    public String getPositionString() {
        if (isEmptyPosition()) {
            return "e";
        } else {
            StringBuilder builder = new StringBuilder();
            for (Integer number : this.basePosition) {
                builder.append(".");
                builder.append(number);
            }
            return builder.substring(1);
        }
    }
    
    
    /**
     * Returns the depth of the position.
     * @return Length of the internal sequence representing the position.
     */
    public int getPositionDepth() {
        return this.basePosition.size();
    }
    
    
    /**
     * Detects whether the position is trivial, i.e. equal to empty sequence.
     * @return Returns true for empty positions.
     */
    public boolean isEmptyPosition() {
        return this.basePosition.isEmpty();
    }


    // -------------------------------------------------------------------------


    /**
     * Prepends a new item at the beginning of the position sequence.
     * @param prefixItem Simple position to be prepended.
     */
    public void applyRootPrepending(Integer prefixItem) {
        assert !isEmptyPosition();
        this.basePosition.add(0, prefixItem);
    }

    /**
     * Increments the first item in the position sequence.
     * @param incrementValue Value to be added to the first simple position.
     */
    public void applyRootShifting(int incrementValue) {
        assert !isEmptyPosition();
        this.basePosition.set(0, this.basePosition.get(0) + incrementValue);
    }

    /**
     * Removes the first item in the position sequence.
     */
    public void applyRootCutting() {
        assert !isEmptyPosition();
        this.basePosition.remove(0);
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates an independent clone of the position operation.
     * @return Newly created independent and identical instance.
     */
    public Position cloneIdenticalPosition() {
        List<Integer> newList = new ArrayList<Integer>();
        for (Integer oldValue : this.basePosition) {
            newList.add(new Integer(oldValue));
        }
        return new Position(newList);
    }

    
    /**
     * Creates a new independent clone with applied root shifting.
     * @param incrementValue Size of the root incrementation.
     * @return Newly created independent position.
     */
    public Position cloneShiftedPosition(int incrementValue) {
        List<Integer> newList = new ArrayList<Integer>();
        for (Integer item : this.basePosition) {
            newList.add(new Integer(item));
        }
        Position newPosition = new Position(newList);
        newPosition.applyRootShifting(incrementValue);
        return newPosition;
    }

    
    // -------------------------------------------------------------------------

}
