
package cz.cuni.mff.corrector.tree;

import java.io.DataOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;

/**
 * Data tree representation for XML documents.
 *
 * @author Martin Svoboda
 */
public class Tree {

    // -------------------------------------------------------------------------


    /**
     * Reference to the root node of the data tree.
     */
    private Node rootNode;


    // -------------------------------------------------------------------------


    /**
     * Creates a new and empty instance of a data tree.
     */
    public Tree() {
        this.rootNode = null;
    }


    /**
     * Creates a new instance of a data tree with a specified root node.
     * @param rootNode Reference to a new root node of a tree.
     */
    public Tree(Node rootNode) {
        this.rootNode = rootNode;
    }


    // -------------------------------------------------------------------------


    /**
     * Creates a new data tree from a given XML document.
     * @param fileName Path to the specified XML file.
     * @return Completely created data tree structure.
     */
    public static Tree createFromXmlFile(String fileName) {
        Node rootNode = XmlParser.parseXmlDocument(fileName);
        Tree dataTree = new Tree(rootNode);
        return dataTree;
    }


    // -------------------------------------------------------------------------


    /**
     * Returns the root node of the data tree.
     * @return Root node reference.
     */
    public Node getRootNode() {
        return this.rootNode;
    }


    /**
     * Recursively counts all nodes in the data tree.
     * @return Overall number of nodes.
     */
    public int countTreeNodes() {
        if (this.rootNode != null) {
            return this.rootNode.countTreeNodes();
        } else {
            return 0;
        }
    }
    
    
    /**
     * Computes the basic statistics about this data tree.
     * @return Container with computed statistics.
     */
    public Stats getTreeStatistics() {
        if (this.rootNode != null) {
            Stats rootStats = this.rootNode.getTreeStatistics();
            if (rootStats.fanoutMin == 0) {
                rootStats.fanoutMin = 1;
            }
            if (rootStats.fanoutMax == 0) {
                rootStats.fanoutMax = 1;
            }
            return rootStats;
        } else {
            return new Stats(0, 0, 0, 0, 0);
        }
    }


    // -------------------------------------------------------------------------


    /**
     * Serializes the data tree into an string.
     * @param prologueStr Optional XML prologue string.
     * @param useIndentation Whether to produce indented document.
     * @return String with serialized data tree.
     */
    public String serializeDataTree(
        String prologueStr,
        boolean useIndentation
    ) {
        if (this.rootNode != null) {
            StringBuilder builder = new StringBuilder();
            if (prologueStr != null) {
                builder.append(prologueStr);
            }
            if (useIndentation) {
                this.rootNode.serializeNode(builder, 0);
            } else {
                this.rootNode.serializeNode(builder);
                builder.append(Node.TOKEN_SE_EOL);
            }
            return builder.toString();
        } else {
            return "";
        }
    }


    /**
     * Exports the given data tree into a new XML file.
     * @param fileName Complete file name of the output file.
     * @param prologueStr Optional XML prologue string.
     */
    public void exportToXmlFile(String fileName, String prologueStr) {
        try {
            FileOutputStream fos = new FileOutputStream(fileName);
            DataOutputStream dos = new DataOutputStream(fos);
            dos.writeBytes(serializeDataTree(prologueStr, true));
        } catch (IOException ex) {
            return;
        }
    }


    // -------------------------------------------------------------------------

}
