
package cz.cuni.mff.corrector.expression;

import org.junit.Test;
import static org.junit.Assert.*;

/**
 * Simple testing class for regular expressions.
 * 
 * @author Martin Svoboda
 */
public class ExpressionTest {

    // -------------------------------------------------------------------------


    /**
     * Tests the correctness of internal function for operand parsing.
     */
    @Test
    public void testOperandParsing() {

        // Definition of input and output pairs
        String[][] container = {
            {"#empty", "|", "#empty"},
            {"#epsilon", "|", "#epsilon"},
            {"a", "|", "a"},
            {"one", ".", "one"},
            {"a.b", "|", "a.b"},
            {"a.b", ".", "a"},
            {"a.b", "*", "a.b"},
            {"a|b", "|", "a"},
            {"a*", "|", "a*"},
            {"a*", "*", "a"},
            {"(a)", "|", "(a)"},
            {"a.b|c", "|", "a.b"},
            {"a.b|c", ".", "a"},
            {"a*.b|c", ".", "a*"},
            {"(a.b|c.d)", "|", "(a.b|c.d)"},
            {"(a)|(b)", "|", "(a)"},
            {"(a.b|c.d)|(d)", "|", "(a.b|c.d)"},
            {"(a*.b*).c*", "|", "(a*.b*).c*"},
            {"(a*.b*).c*", ".", "(a*.b*)"},
            {"((a.b)*)*.c", ".", "((a.b)*)*"},
            {"((((a.b)*)))|(c.d)", "|", "((((a.b)*)))"},
            {"a*.b.c|d*.e|f", "|", "a*.b.c"},
            {"a*.b.c|d*.e|f", ".", "a*"},
            {"a*.b.c|d*.e|f", "*", "a"},
            {"(a)*|(b)*", "|", "(a)*"},
            {"(a)*|(b)*", "*", "(a)"},
        };

        // Testing all prepared expressions
        for (String[] pair : container) {
            int index = Expression.parseOperand(pair[0], pair[1]);
            assertEquals(pair[2], pair[0].substring(0, index));
        }
        
    }


    // -------------------------------------------------------------------------


    /**
     * Tests the correctness of expressions creation and serialization.
     */
    @Test
    public void testExpressionStructure() {

        // Definition of input and output expression pairs
        String[][] container = {
            {"#empty", "#empty"},
            {"#epsilon", "#epsilon"},
            {"a", "a"},
            {"a*|b*.c*", "(a*|(b*.c*))"},
            {"one.two*.(three)", "(one.(two*.three))"},
            {"a|b.c|d", "(a|((b.c)|d))"},
            {"(((a*|b.c)))*|d*", "((a*|(b.c))*|d*)"},
        };

        // Testing all prepared expressions
        for (String[] pair : container) {
            Expression exp = Expression.createExpression(pair[0]);
            assertNotNull(exp);
            String res = exp.serializeExpression();
            assertEquals(pair[1], res);
        }

    }


    // -------------------------------------------------------------------------


    /**
     * Tests the correctness of regular expressions marking.
     */
    @Test
    public void testMarkingExpression() {

        // Definition of input and output expression pairs
        String[][] container = {
            {"#empty", "#empty"},
            {"#epsilon", "#epsilon"},
            {"a", "a#1"},
            {"a*|b*.c*", "(a#1*|(b#2*.c#3*))"},
            {"one.two*.(three)", "(one#1.(two#2*.three#3))"},
            {"a|b.c|d", "(a#1|((b#2.c#3)|d#4))"},
            {"(((a*|b.c)))*|d*", "((a#1*|(b#2.c#3))*|d#4*)"},
        };

        // Testing all prepared expressions
        for (String[] pair : container) {
            Expression exp = Expression.createExpression(pair[0]);
            assertNotNull(exp);
            Expression marked = Expression.markExpression(exp);
            assertNotNull(marked);
            String res = marked.serializeExpression();
            assertEquals(pair[1], res);
        }

    }


    // -------------------------------------------------------------------------
    
}
